rm(list=ls())

library(base);
getParameters = function(parametersPairs, parName, checkSlash=F){
	parameterNames = parametersPairs[,1];
	ind = which(parameterNames==parName);
	parValue = character(0);
	if(length(ind)>0) {
		parValue = parametersPairs[max(ind), 2];
		if(parName == "phenotype_file"){
			parValue = normalizePath(parValue, winslash="/", mustWork=F);
		}
		if(parName == "block1_selected_sites_prefix" || parName == "block2_selected_sites_prefix"){
			dirName = dirname(parValue);
			baseName = basename(parValue);
			parValue = paste(normalizePath(dirName, winslash="/", mustWork=F), "/", baseName, sep="");
		}
		if(checkSlash){
			# convert path to absolute path with slash "/"
			#print(parValue);
			dir.create(parValue, recursive=T, showWarnings = F);
			parValue = normalizePath(parValue, winslash="/", mustWork=F);
			lastChar = substr(parValue, nchar(parValue), nchar(parValue));
			if(lastChar!="/" && lastChar!="\\"){
				parValue = paste(parValue, "/", sep="");
			}
		}
	}
	if(length(parValue)>0){
		cat(paste(parName, "=", parValue, "\n"));
	}
	return(parValue);
}

############### input parameters ######################
Args <- commandArgs(T);

parFile=Args[1];

parametersPairs = read.table(parFile, sep="=", colClasses=c("character", "character"), strip.white=T);

cat("\n#################### Input parameters #############################\n");

cat("#################### Mandatory paramters ##########################\n");
use_cluster=getParameters(parametersPairs, "run_in_parallel");
chr_begin = getParameters(parametersPairs, "chr_begin")
chr_end = getParameters(parametersPairs, "chr_end")
RscriptCmd = getParameters(parametersPairs, "RscriptCmd")
RscriptCmd = normalizePath(RscriptCmd, winslash="/", mustWork=T);
RscriptCmd = paste("\"", RscriptCmd, "\"", sep="");
covdir = getParameters(parametersPairs, "methylation_data_directory", checkSlash=T)
samplefile = getParameters(parametersPairs, "phenotype_file")
#samplefile = normalizePath(samplefile, winslash="/", mustWork=F);
log_dir = getParameters(parametersPairs, "log_directory", checkSlash=T);
cat("###################################################################\n");

cat("\n#################### Create block data ############################\n");
############### level 1 block parameter #################
block1_workdir = getParameters(parametersPairs, "block1_working_directory", checkSlash=T);       
block1_selected_sites_prefix = getParameters(parametersPairs, "block1_selected_sites_prefix");
#block1_selected_sites_prefix = normalizePath(block1_selected_sites_prefix, winslash="/", mustWork=F);
if(length(block1_selected_sites_prefix)==0){
	block1_selected_sites_prefix = "NA";
	block1_selfile = "no";
} else {
	block1_selfile = "yes";
}
block1_min_mean_block_cor = getParameters(parametersPairs, "block1_min_mean_block_cor")   
block1_min_new_block_cor = getParameters(parametersPairs, "block1_min_new_block_cor")
block1_min_n_new_block_cor = getParameters(parametersPairs, "block1_min_n_new_block_cor"); 

############### level 2 block parameter #################
block2_workdir = getParameters(parametersPairs, "block2_working_directory", checkSlash=T);             
block2_selected_sites_prefix = getParameters(parametersPairs, "block2_selected_sites_prefix");
#block2_selected_sites_prefix = normalizePath(block2_selected_sites_prefix, winslash="/", mustWork=F);
if(length(block2_selected_sites_prefix)==0){
	block2_selected_sites_prefix = "NA";
	block2_selfile = "no"
} else {
	block2_selfile = "yes";
}
block2_min_mean_block_cor = getParameters(parametersPairs, "block2_min_mean_block_cor");   
block2_min_new_block_cor = getParameters(parametersPairs, "block2_min_new_block_cor");    
block2_min_n_new_block_cor = getParameters(parametersPairs, "block2_min_n_new_block_cor"); 
cat("###################################################################\n");

cat("\n#################### PCA parameters ###############################\n");
pca_workdir = getParameters(parametersPairs, "pca_working_directory", checkSlash=T)
#covar_file = getParameters(parametersPairs, "covar_file")
pre_pca_indices = getParameters(parametersPairs, "regress_out_covariates_before_pca_indices");
if(length(pre_pca_indices)>0){
	pre_pca_indices = as.integer(strsplit(pre_pca_indices, ",")[[1]]);
}
#resdir = getParameters(parametersPairs, "resdir", checkSlash=T)
desdir= paste(pca_workdir, "descriptives/", sep="");
resdir = paste(pca_workdir, "residuals/", sep="");

n_sample_job=getParameters(parametersPairs, "n_sample_chunk")
if(length(n_sample_job)==0){
	n_sample_job = 0;
}
if(as.integer(n_sample_job) %% 2 == 1){
	n_sample_job = as.character(as.integer(n_sample_job) + 1);
	cat("n_sample_job is changed to the closest even number: n_sample_job\n");
}
n_cores= as.integer(getParameters(parametersPairs, "n_cores"));
if(length(n_cores)==0){
	n_cores = 1;
}
n_parallel_proc=as.integer(getParameters(parametersPairs, "n_parallel_proc"))
memory_per_core = as.numeric(getParameters(parametersPairs, "memory_per_core"));
max_sites = as.numeric(getParameters(parametersPairs, "max_sites"));
max_memory = memory_per_core*n_cores;

loading_type = getParameters(parametersPairs, "loading_type");
number_of_PCs_loadings = getParameters(parametersPairs, "number_of_PCs_loadings");
nPCs = number_of_PCs_loadings;
cat("###################################################################\n");

cat("\n#################### Association test #############################\n");
association_workdir = getParameters(parametersPairs, "association_working_directory", checkSlash=T);
test_label = getParameters(parametersPairs, "test_label");
included_covariates_indices = getParameters(parametersPairs, "included_covariates_indices");
if(length(included_covariates_indices)>0){
	included_covariates_indices = as.integer(strsplit(included_covariates_indices, ",")[[1]]);
}

included_PCs_indices = getParameters(parametersPairs, "included_PCs_indices");
if(length(included_PCs_indices)>0){
	included_PCs_indices = as.integer(strsplit(included_PCs_indices, ",")[[1]]);
	extra_included_covariates_indices = included_PCs_indices;
}
included_PCs_type = getParameters(parametersPairs, "included_PCs_type");
if(length(included_PCs_indices)>0){
	if(length(pca_workdir)==0){
		cat("make sure pca_workdir is set!\n")
		q(save="no")
	}
	if(included_PCs_type == "cov"){
		extra_covariates_file = paste(pca_workdir, "cov_scores.csv", sep="")
	} else if(included_PCs_type == "cor"){
		extra_covariates_file = paste(pca_workdir, "cor_scores.csv", sep="")
	} else {
		cat("wrong value for included_PCs_indices!\n")
		q(save="no");
	}
} else {
	extra_covariates_file = character(0);
}

cat("###################################################################\n");


cat("\n#################### Miscellaneous ################################\n");
job_submission_delay = getParameters(parametersPairs,"job_submission_delay");
validation = getParameters(parametersPairs, "validation")
validation = as.integer(validation);
cat("###################################################################\n");


samplefileData = read.csv(samplefile, header=T, as.is=T);
if(length(pre_pca_indices)>0 && length(pca_workdir) > 0){
	cat("\n#################### Variables regressed out before PCA ###########\n");
	cat(  colnames(samplefileData)[2+pre_pca_indices]);
	cat("\n###################################################################\n");
}
if(length(included_covariates_indices)>0  && length(association_workdir)>0){
	reg_covariates_names = colnames(samplefileData)[2+included_covariates_indices];
	cat("\n#################### Covariates in the association test ###########\n");
	cat(reg_covariates_names);
	cat(" ");
	if(length(included_PCs_indices)>0){
		cat(paste("PC", included_PCs_indices, sep=""));
	}
	cat("\n###################################################################\n");
}


############### some parameter validation here #######
# To do
if(length(block1_workdir)==0 && length(block2_workdir)>0) {
	cat(paste("block1_working_directory is not set while block2_working_directory is set", "\n"));
	q(save="no");
}
##

######################################################
if(Sys.info()['sysname']=="Windows")  {
	pathSep = "/";
} else {
	pathSep = "/";
}

# get the absolute path of this running scipt
initial.options <- commandArgs(trailingOnly = FALSE)
file.arg.name <- "--file="
script.name <- sub(file.arg.name, "", initial.options[grep(file.arg.name, initial.options)])
script.basename <- dirname(script.name)
fullScriptPath = normalizePath(script.basename, winslash="/");



# generate the sampleID file from samplefile
tempdir = log_dir;
if(!file.exists(tempdir)){
	dir.create(tempdir, recursive=T);
}

# generate pre_pca_chr scripts
for(chr in chr_begin:chr_end){
	script_chr = paste(tempdir, "pre_pca_chr", chr, ".bat", sep="") 
	file.create(script_chr);
}

sampleIDs = samplefileData[,1];
sampleIDsFile = paste(log_dir, "sampleIDs.txt", sep="");
cat(sampleIDs, file = sampleIDsFile, sep="\n");
samplefile_orig = samplefile;
samplefile = sampleIDsFile;

# generate the covar_file from the pre_pca_indcie
if(length(pre_pca_indices)>0){
	pre_pca_covariates = samplefileData[, c(1, 2+pre_pca_indices)];
	pre_pca_covariatesFile = paste(log_dir, "pre_pca_covariates.csv", sep="");
	write.csv(pre_pca_covariates, file = pre_pca_covariatesFile, row.names=F, quote=F);
	covar_file = pre_pca_covariatesFile;
}


max_sample = dim(samplefileData)[1];


############################################# Generating blocks ####################################
if(length(block1_workdir)>0 ){ # generating blocks
	# generate level 1 block
	workdir = block1_workdir
	selfile = block1_selfile             
	selected_sites_prefix = block1_selected_sites_prefix
	min_mean_block_cor = block1_min_mean_block_cor
	min_new_block_cor = block1_min_new_block_cor 
	min_n_new_block_cor = block1_min_n_new_block_cor

	if(!file.exists(workdir)){
		dir.create(workdir, recursive=T);
	}
	
	for(chr in chr_begin:chr_end){
		chrDir = paste(workdir, "chr", chr, sep="");
		if(!file.exists(chrDir)){
			dir.create(chrDir, recursive=T);
		}
	
		cmdLine = paste(paste(fullScriptPath, pathSep, "findBlocks.exe", sep=""),  covdir, samplefile, selfile, selected_sites_prefix,
		 workdir,  min_mean_block_cor, min_new_block_cor, min_n_new_block_cor,  chr, "\n");
		script_chr = paste(tempdir, "pre_pca_chr", chr, ".bat", sep="") 
		cat(cmdLine, file=script_chr, append=T);
	}

	
	covdir = block1_workdir; # set the coverage folder 

	
	if(length(block2_workdir)>0 ){
		# generate level 2 block
		workdir = block2_workdir
		selfile = block2_selfile             
		selected_sites_prefix = block2_selected_sites_prefix
		min_mean_block_cor = block2_min_mean_block_cor
		min_new_block_cor = block2_min_new_block_cor 
		min_n_new_block_cor = block2_min_n_new_block_cor

		if(!file.exists(workdir)){
			dir.create(workdir, recursive=T);
		}

		for(chr in chr_begin:chr_end){
			chrDir = paste(workdir, "chr", chr, sep="");
			if(!file.exists(chrDir)){
				dir.create(chrDir, recursive=T);
			}

			cmdLine = paste(paste(fullScriptPath, pathSep, "findBlocks.exe", sep=""), covdir, samplefile, selfile, selected_sites_prefix,
		 workdir,  min_mean_block_cor, min_new_block_cor, min_n_new_block_cor,  chr, "\n");
			script_chr = paste(tempdir, "pre_pca_chr", chr, ".bat", sep="") 
			cat(cmdLine, file=script_chr, append=T);
		}
		
		covdir = block2_workdir; # set the coverage folder
		
		
	}

}
############## end of generating blocks ####################################################################


############################################# perform PCA ##################################################

if(length(pca_workdir)!=0 ){
	
	cov_scores=paste(pca_workdir, "cov_scores.csv", sep="");
	cor_scores=paste(pca_workdir, "cor_scores.csv", sep="");
	cov_eigenval=paste(pca_workdir, "cov_eigenval.csv", sep="");
	cor_eigenval=paste(pca_workdir, "cor_eigenval.csv", sep="");
	cov_eigenvec=paste(pca_workdir, "cov_eigenvec.csv", sep="");
	cor_eigenvec=paste(pca_workdir, "cor_eigenvec.csv", sep="");
	no_pca_results = T;
	
	pca_covdir = covdir;
	
	if(file.exists(cov_scores) && file.exists(cov_scores) && file.exists(cov_scores) && 
		file.exists(cov_scores) && file.exists(cov_scores) && file.exists(cov_scores) && 
		length(block1_workdir) == 0 ) {
		no_pca_results = F;
		cat(paste("Use existing PCA results","\n"));
		
	} else {
	
		workdir = pca_workdir;	
		if(!file.exists(workdir)){
			dir.create(workdir, recursive=T);
		}

		#step1:
		# generate descriptive directory
		
		if(!file.exists(paste(pca_workdir, "descriptives", sep="")) || length(block1_workdir)>0 ){
			dir.create(desdir, recursive=T);
			reuse_des = F;
			
		} else {
			reuse_des = T;
			for (chr in chr_begin:chr_end){
				if( !file.exists(paste(pca_workdir, "descriptives/des_chr", chr, ".csv", sep="")) ){
					print(paste(pca_workdir, "descriptives/des_chr", chr, ".csv", sep=""));
					reuse_des = F;
				}
			}
			# do nothing. Suppose the descriptives and residuals (if regreesing out covariates before PCA) are already computed.
			if(reuse_des){
				cat(paste("Reuse the computed descriptives because descriptives files under subdirectory \"descriptives\" exist", "\n"));
			}
		}

		# no lab covariates
		if (length(pre_pca_indices)==0){
			if(!file.exists(desdir)){
				dir.create(desdir, recursive=T);
			}
			for (chr in chr_begin:chr_end){
				if(reuse_des) {
					cmdLine = "";
				} else {
					cmdLine = paste(paste(fullScriptPath, pathSep, "descriptives.exe", sep=""), covdir, samplefile, desdir, chr, "\n");
				}
				script_chr = paste(tempdir, "pre_pca_chr", chr, ".bat", sep="") 
				cat(cmdLine, file=script_chr, append=T);
			}
		} else {# regressing out lab covariates

			if(!file.exists(resdir)){
				dir.create(resdir, recursive=T);
			}

			for(chr in chr_begin:chr_end){
				chrDir = paste(resdir, "chr", chr, sep="");
				if(!file.exists(chrDir)){
					dir.create(chrDir, recursive=T);
				}
				if(reuse_des) {
					cmdLine = "";
				} else {
					cmdLine = paste(paste(fullScriptPath, pathSep, "calc_residuals_desp.exe", sep=""), covdir, covar_file, resdir, desdir, chr, "\n");
				}
				script_chr = paste(tempdir, "pre_pca_chr", chr, ".bat", sep="") 
				cat(cmdLine, file=script_chr, append=T);
			}

			# change the folder
			pca_covdir=resdir;
		}





		#step2:
		create_innerMa_script = paste(tempdir, "pca_create_scripts_to_compute_XXT.bat", sep="");
		file.create(create_innerMa_script);

		chunksdir=paste(desdir, "chunks_XXT/", sep="");
		if(!file.exists(chunksdir)){
			dir.create(chunksdir, recursive=F);
		}

		# create scripts
		if(n_sample_job <= 0){
			source(paste(fullScriptPath, "/func_partitionXXT2scripts.R", sep=""));
			n_sample_job = func_partitionXXT2scripts(max_sample, max_sites, n_parallel_proc, max_memory);
		}
		cmdLine = paste(RscriptCmd, paste(fullScriptPath, pathSep, "create_PCA_scripts.R", sep=""), paste(fullScriptPath, pathSep, "calc_PCAinput_diag.exe", sep=""), paste(fullScriptPath, pathSep, "calc_PCAinput_rect.exe", sep=""), 
		pca_covdir, samplefile, desdir, tempdir, chunksdir, max_sample, n_sample_job, n_cores, n_parallel_proc, chr_begin, chr_end);
		cat(cmdLine, file=create_innerMa_script);
		


		#step3: 
		outputdir2=desdir;
		assemble_innerMa_script = paste(tempdir, "pca_assemble_XXT.bat", sep="");
		file.create(assemble_innerMa_script);
		cmdLine = paste(RscriptCmd, paste(fullScriptPath, pathSep, "assemble_PCAmatrix.R", sep=""),  chunksdir, samplefile, outputdir2, max_sample, n_sample_job, "\n");
		cat(cmdLine, file=assemble_innerMa_script);

		#step4:
		PCAinput_cov=paste(desdir, "PCAinput_cov.csv", sep="");
		PCAinput_cor=paste(desdir, "PCAinput_cor.csv", sep="");

		calPC_cov_script = paste(tempdir, "pca_cov.bat", sep="");
		file.create(calPC_cov_script);
		cmdLine = paste(paste("cd ", fullScriptPath, pathSep, " &&", sep=""), RscriptCmd, "outputPCs.R", PCAinput_cov, cov_scores, cov_eigenval, cov_eigenvec, "\n");
		cat(cmdLine, file=calPC_cov_script);

		calPC_cor_script = paste(tempdir, "pca_cor.bat", sep="");
		file.create(calPC_cor_script);
		cmdLine = paste(paste("cd ", fullScriptPath, pathSep, " &&", sep=""), RscriptCmd, "outputPCs.R", PCAinput_cor, cor_scores, cor_eigenval, cor_eigenvec, "\n");
		cat(cmdLine, file=calPC_cor_script);

	}
	
	if (length(pre_pca_indices) > 0) {
		pca_covdir=resdir;
	}
	
	#step5: (optional)
	if(length(loading_type)>0 && nPCs>0){
		loading_dir = paste(pca_workdir, "loadings/", sep="");
		if(!file.exists(loading_dir)){
			dir.create(loading_dir, recursive=T);
		}

		if(loading_type == "cov" || loading_type == "both"){
			# calculate loadings: cov
			calLoading_cov_script = paste(tempdir, "pca_loading_cov.bat", sep="");
			file.create(calLoading_cov_script);
			for(chr in chr_begin:chr_end){
				#chrDir = paste(workdir, "chr", chr, sep="");
				#if(!file.exists(chrDir)){
				#	dir.create(chrDir, recursive=T);
				#}
				cmdLine = paste(paste(fullScriptPath, pathSep, "calc_loadings_cov.exe", sep=""), pca_covdir, desdir, pca_workdir, "cov", nPCs, chr, "\n");
				cat(cmdLine, file=calLoading_cov_script, append=T);
			}
		}


		if(loading_type == "cor" || loading_type == "both"){
			# calculate loadings: cor
			calLoading_cor_script = paste(tempdir, "pca_loading_cor.bat", sep="");
			file.create(calLoading_cor_script);

			for(chr in chr_begin:chr_end){
				#chrDir = paste(workdir, "chr", chr, sep="");
				#if(!file.exists(chrDir)){
				#	dir.create(chrDir, recursive=T);
				#}
				cmdLine = paste(paste(fullScriptPath, pathSep, "calc_loadings_cor.exe", sep=""), pca_covdir, desdir, pca_workdir, "cor", nPCs, chr, "\n");			
				cat(cmdLine, file=calLoading_cor_script, append=T);
			}
		}

	}

	# optional: validation

	if(length(validation)>0 && validation == 1 ){
		validation_script = paste(tempdir, "validation_pca.bat", sep="");
		file.create(validation_script)
		validationResult = paste(pca_workdir, "validation_pca.txt", sep="");
		if(length(pre_pca_indices)==0){
			cmdLine = paste(RscriptCmd, paste(fullScriptPath, pathSep, "validation_pca.R", sep=""),  
			covdir, samplefile, pca_workdir, chr_begin, chr_end, ">", validationResult, "\n");
		} else {	# regressing out covariates
			cmdLine = paste(RscriptCmd, paste(fullScriptPath, pathSep, "validation_pca.R", sep=""),  
			covdir, samplefile, pca_workdir, chr_begin, chr_end, covar_file, ">", validationResult, "\n");
		}	
		cat(cmdLine, file=validation_script, append=T);
	}
}

################################################### association test ###############################################
if(length(association_workdir)>0) {
	if(!file.exists(association_workdir)){
		dir.create(association_workdir, recursive=T);
	}

	association_test_script = paste(tempdir, "association_test.bat", sep="");
	file.create(association_test_script);
	
	# generate phenotype files for assocation test
	pheno_file = paste(association_workdir, paste("pheno_file_included_covariates_", test_label, ".csv", sep=""), sep="");
	#pheno_file_included = paste(log_dir, "pheno_file_included.csv", sep="");
	phenoData = samplefileData[,c(1,2)];
	if(length(included_covariates_indices) > 0) {
		covariatesData = samplefileData[, 2+included_covariates_indices];
		pheno_plus_covariates = cbind(phenoData, covariatesData);
	} else {
		pheno_plus_covariates = phenoData;
	}
	write.csv(pheno_plus_covariates, file=pheno_file, quote=F, row.names=F);
			
	if(length(extra_covariates_file)>0){  
		# generate pheno_file with extra covariates
		prepare_pheno_script = paste(tempdir, "association_prepare_pheno.bat", sep="");
		file.create(prepare_pheno_script);
		
		included_covariates_indices_file = paste(tempdir, "included_covariates_indices.txt", sep="");
		write.csv(extra_included_covariates_indices, file=included_covariates_indices_file, row.names=F, quote=F);
		cmdLine = paste(RscriptCmd, paste(fullScriptPath, pathSep, "prepare_pheno.R", sep=""),  
					pheno_file, extra_covariates_file, included_covariates_indices_file, association_workdir, pheno_file, "\n");
		cat(cmdLine, file=prepare_pheno_script, append=T);
	} 
	
	
	for(chr in chr_begin:chr_end){
		cmdLine = paste(paste(fullScriptPath, pathSep, "association_ols.exe", sep=""), 
		covdir, pheno_file, association_workdir, test_label, chr, "\n");
		cat(cmdLine, file=association_test_script, append=T);
	}
	
	
	if(length(validation)>0 && validation == 1 ){
		validation_association_script = paste(tempdir, "validation_association.bat", sep="");
		file.create(validation_association_script)
		validationResult = paste(association_workdir, "validation_association.txt", sep="");
		cmdLine = paste(RscriptCmd, paste(fullScriptPath, pathSep, "validation_association.R", sep=""),  
			covdir, pheno_file, association_workdir, test_label, chr_begin, chr_end, ">", validationResult, "\n");
		cat(cmdLine, file=validation_association_script, append=T);
	}
	
	qqplot_script = paste(tempdir, "association_qqplot.bat", sep="");
	file.create(qqplot_script)
	qqplot_file = paste(association_workdir, "association_", test_label, "_qqplot.png", sep="");
	cmdLine = paste(paste("cd ", fullScriptPath, pathSep, ". &&", sep=""), RscriptCmd, paste(fullScriptPath, pathSep, "association_qqplot.R", sep=""),  
		association_workdir, test_label, chr_begin, chr_end, qqplot_file, "\n");
	cat(cmdLine, file=qqplot_script, append=T); 
}


################################ check the OS and submit jobs #######################
if(Sys.info()['sysname']=="Windows")  {
	if(length(pca_workdir)>0 || length(block1_workdir)>0 || length(block2_workdir)>0){
		for(chr in chr_begin:chr_end){
			script_chr = paste(tempdir, "pre_pca_chr", chr, ".bat", sep="") 
			system(script_chr);	
		}
	}
	
	if( length(pca_workdir)>0 ){
		
		if(no_pca_results){
			system(create_innerMa_script);
			for (i in 1:n_parallel_proc){
				#print(paste(tempdir,"pca_compute_XXT_",i,".bat", sep=""));
				system(paste(tempdir,"pca_compute_XXT_",i,".bat", sep=""));
			}
			system(assemble_innerMa_script);
			system(calPC_cov_script);
			system(calPC_cor_script);
		}
		
		if(length(loading_type)>0 && nPCs>0){
			if(loading_type == "cov" || loading_type == "both"){
				system(calLoading_cov_script);
			}
			if(loading_type == "cor" || loading_type == "both"){
				system(calLoading_cor_script);
			}
		}
	
		if(length(validation)>0 && validation == 1){
			system(validation_script);
		}
	}
	
	if(length(association_workdir)>0) {
		if(length(extra_covariates_file)>0){
			system(prepare_pheno_script);
		}
		system(association_test_script);
		system(qqplot_script);
		
		if(length(validation)>0 && validation == 1){
			system(validation_association_script);
		}
	}
	
} else if (use_cluster==0){  # Linux/Unix based OS
	if(length(pca_workdir)>0 || length(block1_workdir)>0 || length(block2_workdir)>0){
		for(chr in chr_begin:chr_end){
			script_chr = paste(tempdir, "pre_pca_chr", chr, ".bat", sep="") 
			system(paste("bash", script_chr));	
		}
	}
	
	if( length(pca_workdir)>0 ){
		if(no_pca_results){
			system(paste("bash", create_innerMa_script));
			for (i in 1:n_parallel_proc){
				system(paste("bash", paste(tempdir,"pca_compute_XXT_",i,".bat", sep="")));
			}
			system(paste("bash", assemble_innerMa_script));

			system(paste("bash", calPC_cov_script));
			system(paste("bash", calPC_cor_script));
		
		}
		
		if(length(loading_type)>0 && nPCs>0){
			if(loading_type == "cov" || loading_type == "both"){
				system(paste("bash", calLoading_cov_script));
			}
			if(loading_type == "cor" || loading_type == "both"){
				system(paste("bash", calLoading_cor_script));
			}
		}
		if(length(validation)>0 && validation == 1 ){
			system(paste("bash", validation_script));
		}
	}
	
	if(length(association_workdir)>0) {
		if(length(extra_covariates_file)>0){
			system(paste("bash", prepare_pheno_script));
		}
		system(paste("bash", association_test_script));
		system(paste("bash", qqplot_script));
		if(length(validation)>0 && validation == 1){
			system(paste("bash", validation_association_script));
		}
	}
	
} else { # use the cluster under Linux/Unix environment
	if(length(job_submission_delay)>0){
		#print(job_submission_delay);
		timeGap = as.integer(job_submission_delay);
	} else {
		timeGap = 0;
	}
	tempscriptFile = paste(tempdir, "qsub_script.sh", sep="");
	logPath = tempdir;
	qsubName = "";
	
	source(paste(fullScriptPath, pathSep, "qsubCmdLinesInFile.R", sep=""));
	if(length(pca_workdir)>0 || length(block1_workdir)>0 || length(block2_workdir)>0){
		cat("Submitting pre-pca jobs ...\n")
		depName = qsubName; qsubName = "pre_pca";
		for(chr in chr_begin:chr_end){
			script_chr = paste(tempdir, "pre_pca_chr", chr, ".bat", sep="")
			qsubCmdLinesInFile(script_chr, eachLine=F, qsubName, depName, timeGap, tempscriptFile, logPath, 0, 0, chr);
		}
	}
	
	if( length(pca_workdir)>0 ){
		cat("Submitting pca jobs\n");
		if(no_pca_results){
			system(paste("bash", create_innerMa_script));
			#depName = qsubName; qsubName = "pca_scrp";
			#qsubCmdLinesInFile(create_innerMa_script, eachLine=F, qsubName, depName, timeGap, tempscriptFile, logPath, 0, 0);
			depName = qsubName; qsubName = "calcuXXT";
			for (i in 1:n_parallel_proc){
				qsubCmdLinesInFile(paste(tempdir,"pca_compute_XXT_",i,".bat", sep=""), eachLine=F, qsubName, depName, timeGap, tempscriptFile, logPath, n_cores, max_memory, i);
			}
			qsubCmdLinesInFile(assemble_innerMa_script, eachLine=T, "assemXXT", "calcuXXT", timeGap, tempscriptFile, logPath, 0, 0);
			qsubCmdLinesInFile(calPC_cov_script, eachLine=T, "calPCs", "assemXXT", timeGap, tempscriptFile, logPath, 0, 0);
			qsubCmdLinesInFile(calPC_cor_script, eachLine=T, "calPCs", "assemXXT", timeGap, tempscriptFile, logPath, 0, 0);
		}
		
		if(length(loading_type)>0 && nPCs>0){
			if(loading_type == "cov" || loading_type == "both"){
				qsubCmdLinesInFile(calLoading_cov_script, eachLine=T, "calLdCov", "calPCs", timeGap, tempscriptFile, logPath, 0, 0);
			}
			if(loading_type == "cor" || loading_type == "both"){
				qsubCmdLinesInFile(calLoading_cor_script, eachLine=T, "calLdCor", "calPCs", timeGap, tempscriptFile, logPath, 0, 0);
			}
		}

		if(length(validation)>0 && validation == 1 ){
			depName = "calLdCor"; qsubName = "validate";
			qsubCmdLinesInFile(validation_script, eachLine=T, qsubName, depName, timeGap, tempscriptFile, logPath, 0, 0);
		}
	}
	
	if(length(association_workdir)>0) {
		cat("Submitting association test jobs\n");
		depName = "calPCs"; qsubName = "prepPhen";
		if(length(extra_covariates_file)>0){
			qsubCmdLinesInFile(prepare_pheno_script, eachLine=T, qsubName, depName, timeGap, tempscriptFile, logPath, 0, 0);
		}
		depName = qsubName; qsubName = "associate";
		qsubCmdLinesInFile(association_test_script, eachLine=T, qsubName, depName, timeGap, tempscriptFile, logPath, 0, 0);
		qsubCmdLinesInFile(qqplot_script, eachLine=T, "qqplot", "associate", timeGap, tempscriptFile, logPath, 0, 0);
		
		if(length(validation)>0 && validation == 1 ){
			depName = qsubName; qsubName = "validate";
			qsubCmdLinesInFile(validation_association_script, eachLine=T, qsubName, depName, timeGap, tempscriptFile, logPath, 0, 0);
		}
	}
}

#################################################################################

