#include <cstdlib>

#include <iostream>
#include <fstream>
#include <string>
#include <sstream>

#include <vector>
#include <cmath>

#include <stdio.h>

#include "constants.h"

using namespace std;

string sample_names[max_sample];

double eigen_val[max_sample], constant[max_sample];
double mean_vector[max_coord], sd_vector[max_coord];

typedef vector<vector<double> > matrix_t;
matrix_t eigen_vec, loadings;
void sizeMatrix(matrix_t &, int,int);

string eigenfiles_label;
int n_sample, n_pcs;

void read_n_sample(const char * filename);
void read_eigenvec(const char * filename, int n_pcs_copy);
void read_eigenval(const char * filename);
void read_des(const char * filename);



int main(int argc, char *argv[])
{

    #ifdef __WIN__
	  // Windows stuff
	  _setmaxstdio(2048);
	#endif

    eigenfiles_label = argv[4];
    int chrom = atoi(argv[6]);
    int n_pcs = atoi(argv[5]);

	string workdir = argv[3];

   string covdir    = argv[1];
   string desdir    = argv[2];

    cout << " Calculate loadings for the first " << n_pcs << " principal components \n";

    int i, j, sample_nr;
    string STRING, subst, filna;
	size_t pos, oldpos;


    stringstream ss;
	const char *ptemp;


	// read the eigenvalues and eigenvectors
    ss << workdir << eigenfiles_label << "_eigenval.csv";
    filna = ss.str();
    ss.str(string());
    ptemp= filna.c_str();
    read_eigenval(ptemp);



    ss << workdir << eigenfiles_label << "_eigenvec.csv";
    filna = ss.str();
    ss.str(string());
    ptemp= filna.c_str();
    read_n_sample(ptemp);
    cout << " Number of samples is " <<  n_sample << "\n";

    sizeMatrix(eigen_vec, n_sample,n_pcs);
    read_eigenvec(ptemp,n_pcs);


    ifstream desfil;
    ss << desdir << "des_chr" << chrom << ".csv";
    filna.clear();
    filna = ss.str();
    ss.str(string());
    ptemp= filna.c_str();
    (desfil).open(ptemp);
    if (!(desfil).is_open()) {cout << "The file " << ptemp << "  is not open \n"; exit (1);}
    cout << "file named " << ptemp << "  is read \n";

    read_des(ptemp);


    // Open coverage files
    ifstream infils[n_sample];
    double cpgdata[n_sample];
    for(i=0;i<n_sample;i++) {
        ss << covdir << "chr" << chrom << "/" << sample_names[i] << ".csv" ;
        filna.clear();
        filna = ss.str();
        ss.str(string());
        ptemp= filna.c_str();
        (infils[i]).open(ptemp);
        if (!(infils[i]).is_open()) {cout << "The file " << ptemp << "  is not open \n"; exit (1);}
        }
     for(i=0;i<n_sample;i++) getline(infils[i],STRING); // Skipping the headers


     for(j=0;j<n_pcs;j++)  constant[j] = 1/(sqrt(n_sample-1) * sqrt(eigen_val[j])) ;
     sizeMatrix(loadings, max_coord,n_pcs);
     int i_coord=0,n_coord=0;
     double temp_value = 0;

     while( !infils[0].eof() ) {    //  Here starts the loop over coordinates

       for(i=0;i<n_sample;i++) {
	     	       getline(infils[i],STRING);
	              oldpos = STRING.find(',');
	              pos=STRING.find(',',oldpos+1);
	              subst = STRING.substr(oldpos+1,pos-oldpos-1);
	              if (sd_vector[i_coord] > sd_threshold) {
					  cpgdata[i] = (atof(subst.c_str()) - mean_vector[i_coord]) / sd_vector[i_coord];
	              } else {
					  cpgdata[i] = 0;
					  if(i==0) {
						  cout << " Standardard deviation of site " << i_coord << " is (very close to) zero: Loadings set to zero\n";
					  }
				  }
       }


          for(j=0;j<n_pcs;j++) {
              temp_value = 0;
              for(i=0;i<n_sample;i++) temp_value =  temp_value + cpgdata[i]*eigen_vec[i][j]*constant[j];
              loadings[i_coord][j] = temp_value;
              }
        i_coord++;

     } // end while

    for(i=0;i<n_sample;i++) (infils[i]).close();

    if (STRING == "") n_coord = i_coord-1; else n_coord = i_coord ; // correct in case of empty final line

    // Write results
    ss << workdir << "loadings/" << eigenfiles_label  << "_loadings" << "_chr" << chrom << ".csv";
    filna = ss.str();
    ss.str(string());
    ptemp= filna.c_str();
    cout << " The name of the output file is " << ptemp << "\n";

    ofstream outfile;
    outfile.open(ptemp);
    outfile.precision(15);
    outfile << "block"; // Creating header
    for(j=0;j<n_pcs;j++) outfile << ",pc" << j+1;
    outfile << "\n";
    for(i=0;i<n_coord;i++) {
        outfile << i+1;
        for(j=0;j<n_pcs;j++) outfile << "," <<  loadings[i][j];
        outfile << "\n";
    }


    outfile.close();

    return EXIT_SUCCESS;
}

//---------functions----------------

void read_des(const char * filename) {

    string STRING, subst;
	ifstream infile;
	size_t pos, oldpos, pos1, pos2;
    int i=0;

	infile.open (filename);

	getline(infile,STRING); // Skipping the header

    while(!infile.eof()) { // To get you all the lines.

        getline(infile,STRING);
        if (STRING=="") break;  // ++ by chen
	oldpos = STRING.find(',');
        pos=STRING.find(',',oldpos+1);
        subst = STRING.substr(oldpos+1,pos-oldpos-1);
        mean_vector[i] = atof(subst.c_str());

        oldpos = pos + 1;
        pos=STRING.find(',',oldpos);
        subst = STRING.substr(oldpos,pos-oldpos);
        sd_vector[i] = atof(subst.c_str());


        i++;
    }

    infile.close();

}

void read_eigenval(const char * filename) {

    string STRING;
	ifstream infile;
	int i=0;

	infile.open (filename);
    cout << " file " << filename << " with eigenvalues is read \n";
    if (!(infile).is_open()) { cout << " eigenvalue file " << filename << "  is not open \n"; exit(-1); }

    getline(infile,STRING); // skip header

    while(!infile.eof()) {
           getline(infile,STRING);  // ++ by chen
           if (STRING=="") break;   // ++ by chen
           eigen_val[i]= atof(STRING.c_str()); // ++ by chen
	   i++;
    }


    infile.close();
}



void read_n_sample(const char * filename) {

    string STRING;
	ifstream infile;
	int i=0;

	infile.open (filename);
    cout << " file " << filename << " with eigenvectors is read \n";
    if (!(infile).is_open()) { cout << " eigenvector file " << filename << "  is not open \n"; exit(-1);}

	getline(infile,STRING); // Skip header

     while(!infile.eof()) {
	   getline(infile,STRING);
	   if (STRING=="") break;
	   i++;
    }
    n_sample = i;
    infile.close();
}

void read_eigenvec(const char * filename, int n_pcs_copy) {

    string STRING, subst, subst2;
	ifstream infile;
	size_t pos, oldpos, pos1, pos2;
	int i=0, j;

	infile.open (filename);
    cout << " file " << filename << " with eigenvectors is read \n";
    if (!(infile).is_open()) { cout << " eigenvector file " << filename << "  is not open \n"; exit(-1); }

	getline(infile,STRING); // Skip header

     while(!infile.eof()) {
	   getline(infile,STRING);
	    if (STRING=="") break;
	   oldpos = 0;
	   pos=STRING.find(',');
	   if (pos==string::npos) break;
       subst = STRING.substr(oldpos,pos-oldpos);
       pos1 = subst.find('"');
       pos2 = subst.find('"',pos1+1);
       subst2 = subst.substr(pos1+1,pos2-pos1-1); // getting rid of "
       sample_names[i] = subst2;   // creating the names of all coverage files

       for(j=0;j<n_pcs_copy;j++) {
           oldpos = pos+1;
           pos=STRING.find(',',oldpos);
           subst = STRING.substr(oldpos,pos-oldpos);
           eigen_vec[i][j]  = atof(subst.c_str());
            }
       i++;
    }

    infile.close();
}


void sizeMatrix(matrix_t & m , int r, int c)
{
  m.clear();
  m.resize(r);
  for (int i=0; i<r; i++)
    m[i].resize(c,0);
}


