#include <cstdlib>

#include <iostream>
#include <fstream>
#include <string>
#include <sstream>

#include <vector>
#include <cmath>
#include <ctime>
#include <stdio.h>


// gsl
#include <gsl/gsl_fit.h>
#include <gsl/gsl_multifit.h>
#include <gsl/gsl_cdf.h>

#include "constants.h"

using namespace std;

string covar_filenames[max_sample];
double res_score[max_sample];

int coord[max_coord];
double mean_resid[max_coord];
double SD_resid[max_coord];

typedef vector<vector<double> > matrix_t;
matrix_t covar_data;
void sizeMatrix(matrix_t &, int,int);

double chisq, pred_score;

string covar_file;
int n_sample, n_coord, n_cov;

void gsl_calc_resid_wrapper (double* cpgdata, \
                                  matrix_t& covariates, int n_covariates, \
                                  int n_sample );

void read_covar(const char * filename);
void covar_matrix_dimensions(const char * filename);
void read_coord(const char * filename);


int main(int argc, char *argv[])
{
	#ifdef __WIN__
	  // Windows stuff
	  _setmaxstdio(2048);
	#endif

 	// ++ by chen
 	covar_file = argv[2];
	int ichrom = atoi(argv[5]);

	string workdir = argv[4];

	string covdir = argv[1];
 	string resdir = argv[3];

    struct tm *current;
	time_t now;
    time(&now);
	current = localtime(&now);
    cout << " Start at " << current->tm_hour << ":" << current->tm_min << ":" << current->tm_sec << "\n";
    cout << " \n";

    cout << " The chromosome being processed is " << ichrom << "\n";
    cout << " \n";

    int i, j, iter;
    string STRING, subst, filna;
	size_t pos, oldpos;

    stringstream ss;
	const char *ptemp;

    double mean=0.0, SD=0.0, denom=0.0, aux=0.0;


	// read the file with sample IDs and covariates that will be regresseed put
  	ifstream infil;
    // ss << workdir << covar_file; // -- by chen
    ss << covar_file;
    filna = ss.str(); ss.str(string()); ptemp= filna.c_str(); (infil).open(ptemp);
    if (!(infil).is_open()) {cout << "The file with covariates " << ptemp << "  is not open \n"; exit (1);}
       else cout << " The name of the file with covariates " <<  ptemp  << "\n";
    cout << " \n";
    covar_matrix_dimensions(ptemp);  // n_sample and n_cov are determined
    cout << " The number of samples is " << n_sample << "\n";
    cout << " \n";
    cout << " The number of covariates is " << n_cov << "\n";
    cout << " \n";

    sizeMatrix(covar_data, n_sample,n_cov);
    read_covar(ptemp);

     // Determine coordinate length from 1st coverage file and open all coverage files
    ifstream infils[n_sample];
    i = 0;
    ss << covdir << "chr" << ichrom << "/" << covar_filenames[i];
    filna.clear(); filna = ss.str(); ss.str(string()); ptemp= filna.c_str();
    (infils[i]).open(ptemp);
    if (!(infils[i]).is_open()) {cout << "file named " << ptemp << "  is not open \n"; exit (1);}
    read_coord(ptemp);
    cout << " The number of coordinates is " << n_coord  << "\n";
    cout << " \n";
    (infils[i]).close();

  	for(i=0;i<n_sample;i++) {
       ss << covdir << "chr" << ichrom << "/" << covar_filenames[i];
       filna.clear(); filna = ss.str(); ss.str(string()); ptemp= filna.c_str();
       (infils[i]).open(ptemp);
       if (!(infils[i]).is_open()) {cout << " File for residual scores " << ptemp << "  is not open \n"; exit (1);}
        }
    cout << " Final file with residual scores is " << filna << "\n";
    cout << " \n";


    for(i=0;i<n_sample;i++) getline(infils[i],STRING); // Skipping the headers


    // Open files to write residual scores
    ofstream outfils[n_sample];
  	for(i=0;i<n_sample;i++) {
  //      ss << resdir << "chr" << ichrom << "/" << covar_filenames[i]; // --  by chen
        ss << resdir << "chr" << ichrom << "/" << covar_filenames[i]; // uncommented by chen
       filna.clear(); filna = ss.str(); ss.str(string()); ptemp= filna.c_str(); (outfils[i]).open(ptemp);
       outfils[i].setf(ios::scientific); outfils[i].precision(15);
       if (!(outfils[i]).is_open()) {cout << " Output file with residuals " << ptemp << "  is not open \n"; exit (1);}
        }
    for(i=0;i<n_sample;i++)  outfils[i] << "blocknr,coverage" << "\n"; // Creating header


    double cpgdata[n_sample];
    // Here starts the loop on coordinates;
    for(iter=0;iter<n_coord;iter++) {

       for(i=0;i<n_sample;i++) {
  	      getline(infils[i],STRING);
          oldpos = STRING.find(',');
          pos=STRING.find(',',oldpos+1);
          subst = STRING.substr(oldpos+1,pos-oldpos-1);
          cpgdata[i] = atof(subst.c_str());
        }

       // calculate the residual scores
       gsl_calc_resid_wrapper (cpgdata, \
                               covar_data, n_cov, \
                               n_sample );


       // write residuals + calculate their mean
       mean=0.0, SD=0.0, aux=0.0, denom=0.0;
       for(i=0;i<n_sample;i++) {
              outfils[i] << coord[iter] << "," << res_score[i] <<  "\n";
              mean +=res_score[i];;
           }

       mean /= n_sample;
       for(i=0;i<n_sample;i++) {
          aux = res_score[i] - mean;
          SD += aux*aux;  }
       SD /= (n_sample-1);
       if (SD > sd_threshold) SD = sqrt(SD); else SD=0.0;                         ;

       mean_resid[iter] = mean;
       SD_resid[iter] = SD;


     }  // end   for(iter=0;iter<n_coord;iter++)


    for(i=0;i<n_sample;i++) {
              (infils[i]).close();
              (outfils[i]).close();  }

    ofstream outfile;
    ss << workdir << "des_chr" << ichrom << ".csv";  // -- "test" by chen
    filna = ss.str(); ss.str(string()); ptemp= filna.c_str(); outfile.open(ptemp);
    if (!(outfile).is_open()) {cout << " Output file with descriptives " << ptemp << "  is not open \n"; exit (1);}
    outfile.setf(ios::scientific); outfile.precision(15);
    cout << " The name of the output file with descriptives is " << filna << "\n";
    outfile << "coord,mean,SD" << "\n"; // Creating header
    for(i=0;i<n_coord;i++) outfile << coord[i] << "," << mean_resid[i] << "," << SD_resid[i] << "\n";

    outfile.close();


   time(&now);
   current = localtime(&now);
   cout << " \n";
   cout << " Finish at " << current->tm_hour << ":" << current->tm_min << ":" << current->tm_sec << "\n";


    return EXIT_SUCCESS;
}

//---------functions----------------

void gsl_calc_resid_wrapper (double* cpgdata, \
                                  matrix_t& covariates, int n_covariates, int n_sample ) {
    gsl_matrix *X, *cov;
    gsl_vector *y, *w, *c;

    int n = n_sample;
    int p = 1 + n_covariates; // intercept + covariates
    // allocate the memory
    X = gsl_matrix_alloc (n, p);
    y = gsl_vector_alloc (n);
    w = gsl_vector_alloc (n);

    c = gsl_vector_alloc (p);
    cov = gsl_matrix_alloc (p, p);
    // setup the data
    for (int i=0; i<n; i++){
        // fill X
        gsl_matrix_set(X, i, 0, 1.0);
        for (int j=0; j<n_covariates; j++){
            gsl_matrix_set(X, i, 1+j, covariates[i][j]);
        }
        // fill y
        gsl_vector_set(y, i, cpgdata[i]);
      }

    // invoke the gls function for weighted multiple regression
    {
        gsl_multifit_linear_workspace * work \
               = gsl_multifit_linear_alloc (n, p);

        gsl_multifit_linear (X, y, c, cov, \
                                   &chisq, work);
        gsl_multifit_linear_free (work);
    }



    // calculate the residuals
    #define C(i) (gsl_vector_get(c,(i)))
    for  (int i=0; i<n; i++) {
       pred_score =  C(0);
       for  (int j=0; j<n_covariates; j++) {
          pred_score = pred_score +  C(j+1)*covariates[i][j]; }
       res_score[i] = cpgdata[i] - pred_score;
     }

    // free the memory
    gsl_matrix_free (X);
    gsl_vector_free (y);
    gsl_vector_free (w);
    gsl_vector_free (c);
    gsl_matrix_free (cov);

}



void covar_matrix_dimensions(const char * filename) {

    string STRING, subst, subst2;
	ifstream infile;
	size_t pos, oldpos, pos1, pos2;
	int i=0, j;

	infile.open (filename);

	getline(infile,STRING); // Determine number of covariates
	oldpos = 0;
	pos=STRING.find(',');
    j=0;
    while (pos!=string::npos) {
       oldpos = pos+1;
       pos=STRING.find(',',oldpos);
       subst = STRING.substr(oldpos,pos-oldpos);
       j++;
    }
    n_cov = j ;

    while(!infile.eof()) {
	        getline(infile,STRING);
            i++;
    }

    n_sample = i -1;
    infile.close();
}


void read_covar(const char * filename) {

    string STRING, subst, subst2;
	ifstream infile;
	size_t pos, oldpos, pos1, pos2;
	int i=0, j;

	infile.open (filename);
    if (!(infile).is_open()) {cout << "The file " << filename << "  is not open \n"; exit (1);}
	getline(infile,STRING); // Skip header

    while(!infile.eof()) {
	   getline(infile,STRING);
	   oldpos = 0;
	   pos=STRING.find(',');
	   if (pos==string::npos) break;
       subst = STRING.substr(oldpos,pos-oldpos);
       //pos1 = subst.find('"');
       //pos2 = subst.find('"',pos1+1);
       //subst2 = subst.substr(pos1+1,pos2-pos1-1); // getting rid of "
       covar_filenames[i] = subst + ".csv";   // creating the names of all coverage files
	   //cout<<covar_filenames[i]<<endl;

       j= 0;
      while (pos!=string::npos) {
           oldpos = pos+1;
           pos=STRING.find(',',oldpos);
           subst = STRING.substr(oldpos,pos-oldpos);
           covar_data[i][j]  = atof(subst.c_str());
           j++;
            }

       i++;
    }

    infile.close();
}



void read_coord(const char * filename) {

	ifstream infile;
    string STRING;

	infile.open (filename);

    getline(infile,STRING);

   	int i=0;
    while(!infile.eof()) {
          getline(infile,STRING);
          if (STRING=="") break;
          coord[i] = atoi(STRING.c_str());
          i++;
    }

    n_coord = i;

    infile.close();
}




void sizeMatrix(matrix_t & m , int r, int c)
{
  m.clear();
  m.resize(r);
  for (int i=0; i<r; i++)
    m[i].resize(c,0);
}


