#include <cstdlib>

#include <iostream>
#include <fstream>
#include <string>
#include <sstream>

#include <vector>
#include <cmath>
#include <ctime>

#include <stdio.h>

#include "constants.h"

using namespace std;


const int max_block_size = 10000;      // max number of CpG sites in a block

double missing_code = -1.0;    // numeric code for missing values


string sample_names[max_sample];

int coord[max_coord], select_sites[max_coord];

int block_start_i_cpg[max_coord], block_end_i_cpg[max_coord], block_size_select[max_coord];
double block_inter_cor[max_coord ];

typedef vector<vector<double> > matrix_double_dynamic;
matrix_double_dynamic cpg_data_matrix;
void shape_matrix(matrix_double_dynamic &, int,int);

void read_sample(const char * filename, int& nr_sample );
void read_select_sites(const char * filename,  int& nr_cpg);
void read_first_covfil(const char * filename,  int& nr_cpg);

double calc_cor(double * x, double * y, int n);
double calc_mean_cor(matrix_double_dynamic & X, double * y, int m, int n);

int main(int argc, char *argv[])
{
	#ifdef __WIN__
	  // Windows stuff
	  _setmaxstdio(2048);
	#endif

   int ichrom = atoi(argv[9]);
   double min_mean_block_cor  = atof(argv[6]);  // threshold for average inter block correlation
   double min_new_block_cor   = atof(argv[7]);  // threshold for correlation of new site vs sites already in block
   int    min_n_new_block_cor = atoi(argv[8]);  // will stop expanding block if min_n_new_block_cor new sites have correlation below min_new_block_cor with sites already in block

   string selfile     = argv[3]; // will you use file to select sites from this chrosomosme
   string covdir      = argv[1]; // root dirtectory with coverage files
   string samplefil   = argv[2]; // file with sample IDs
   string selected_cites_prefix = argv[4]; // ++ by chen

    cout << "\n";
    cout << " The chromosome being processed is " << ichrom << "\n";
    cout << " Threshold for average inter block correlation (min_mean_block_cor) is " << min_mean_block_cor << "\n";
    cout << " Threshold for correlation of new site vs sites already in block (min_new_block_cor) is " << min_new_block_cor  << "\n";
    cout << " Stop expanding block if # new sites have low correlation with sites already in block (min_n_new_block_cor) is " << min_n_new_block_cor << "\n";
    cout << "\n";

    cout << " Will a file to select sites be used (if not all sites will be used): " << selfile  << "\n";
    cout << " Root directory for coverage files is " << covdir << "\n";
    cout << " File with sample IDs is " << samplefil << "\n";
    cout << "\n";

   string workdir   = argv[5];

   struct tm *current;
	time_t now;
    time(&now);
	current = localtime(&now);
    cout << " Start at " << current->tm_hour << ":" << current->tm_min << ":" << current->tm_sec << "\n";
    cout << " \n";


    cout << " The chromosome being processed is " << ichrom << "\n";
    cout << " \n";

    int i, j;

    stringstream ss;
	const char *ptemp;
    string STRING, subst, filna;
	size_t pos, oldpos;

    // read sample names and determine actual samples size

    int n_sample=0;
    ss << samplefil;
    filna = ss.str(); ss.str(string()); ptemp= filna.c_str();
    read_sample(ptemp,n_sample);
    cout << " The number of samples in " << ptemp << " is " << n_sample << "\n";
    cout << " \n";

  	ifstream infils[n_sample];
    int n_cpg = 0;

    if (selfile=="yes") {
        //ss << workdir << "select_sites/sel_chr" << ichrom << ".csv";   // -- by chen
        ss << selected_cites_prefix << ichrom << ".csv";
	filna = ss.str(); ss.str(string()); ptemp= filna.c_str();
	read_select_sites(ptemp,n_cpg);
        cout << " The name of the site selection file is " << filna << "\n";
        cout << "    ...number of sites in this file is " << n_cpg << "\n";
        cout << " \n";
     } else { // read frust coverage file to determone the number of sites
          ss << covdir << "chr" << ichrom << "/" << sample_names[0] + ".csv";
          filna.clear(); filna = ss.str(); ss.str(string()); ptemp= filna.c_str(); (infils[0]).open(ptemp);
          read_first_covfil(ptemp,n_cpg);
          cout << " No selection of sites\n";
          cout << "    ...number of sites according to coverage files is " << n_cpg << "\n";
          cout << " \n";
       }
    (infils[0]).close();

   	for(i=0;i<n_sample;i++) {
       ss << covdir << "chr" << ichrom << "/" << sample_names[i] + ".csv";
       filna.clear(); filna = ss.str(); ss.str(string()); ptemp= filna.c_str(); (infils[i]).open(ptemp);
       if (!(infils[i]).is_open()) {cout << " CpG data file number " << i << " named " << ptemp << "  is not open \n"; exit(1);}
       getline(infils[i],STRING);              // skipping the headers if input files
    }
    cout << " Final methylation data file read is " << filna << "\n";
    cout << " \n";

    ofstream outfils[n_sample];
  	for(i=0;i<n_sample;i++) {
        ss << workdir << "chr" << ichrom << "/" << sample_names[i] + ".csv"; // test removed by chen
        filna.clear(); filna = ss.str(); ss.str(string()); ptemp= filna.c_str(); (outfils[i]).open(ptemp);
        outfils[i].setf(ios::scientific); outfils[i].precision(15);
        if (!(outfils[i]).is_open()) {cout << " Block data output file number " << i << " named " << ptemp << "  is not open \n"; exit(1);}
        outfils[i] << "coord,coverage" << "\n"; // creating headers for output files
    }
    cout << " Final block data output file opened " << filna << "\n";
    cout << " \n";

    shape_matrix(cpg_data_matrix, max_block_size, max_sample);

    double cpg_data[n_sample];
    double sum_cpg_data[n_sample];
    double new_cpg_data[n_sample];

    int block_start = 0, block_size = 2, n_new_block_cor = 0, i_cpg = 0, i_cpg_old =0, first_cpg = 0, i_block = 0, prev_block_size=0;
    double mean_cor = 1.0, prev_mean_cor = 0.0, n_cor_double = 1.0, weight = 0.0, new_block_cor=0.0;

//    n_cpg = 10000;
    while (i_cpg < n_cpg) {

        mean_cor = 1.0, prev_mean_cor = 0.0; block_size = 2, n_new_block_cor  = 0;

        while ( select_sites[i_cpg]==0 ) {   //skipping CPGs that are excluded
                for(i=0;i<n_sample;i++) getline(infils[i],STRING);
                 i_cpg++;
        }
        if (first_cpg==0) {   // need to read 1st CpG data, after this it will take the last read CpG data that could not be combined in previuos block
            for(i=0;i<n_sample;i++) {
               getline(infils[i],STRING);
               oldpos = STRING.find(',');
               pos=STRING.find(',',oldpos+1);
               subst = STRING.substr(oldpos+1,pos-oldpos-1);
               sum_cpg_data[i] = atof(subst.c_str());
             }
        first_cpg=1;
        } else for(i=0;i<n_sample;i++) sum_cpg_data[i] = new_cpg_data[i];  // starting with the last data that could not be combined

        block_start = i_cpg;  // now read next CpG
        i_cpg++;
        while ( select_sites[i_cpg]==0 ) { // skip until next selected CpG
                  for(i=0;i<n_sample;i++) getline(infils[i],STRING);
                  i_cpg++;
        }
        for(i=0;i<n_sample;i++) {
       	    getline(infils[i],STRING);
            oldpos = STRING.find(',');
            pos=STRING.find(',',oldpos+1);
            subst = STRING.substr(oldpos+1,pos-oldpos-1);
            new_cpg_data[i] = atof(subst.c_str());
        }

        mean_cor = calc_cor(sum_cpg_data,new_cpg_data,n_sample);

        // write if we cannot combine, if we can combine store data in cpg_data_matrix to see if we can add further sites
//         if ( (mean_cor < min_mean_block_cor) && ( block_size == 2) && select_sites[i_cpg] == 1  )  {
         if ( mean_cor < min_mean_block_cor )  {
            block_start_i_cpg[i_block] = block_start;
            block_end_i_cpg[i_block]   = block_start;
            block_inter_cor[i_block]   = missing_code;
            block_size_select[i_block] = 1;

            for(i=0;i<n_sample;i++) outfils[i] << i_block+1 << "," << sum_cpg_data[i] << "\n";

            i_block++;

            }  else for(j=0;j<n_sample;j++) { cpg_data_matrix[0][j] = sum_cpg_data[j]; cpg_data_matrix[1][j] = new_cpg_data[j]; }


        while ((mean_cor > min_mean_block_cor) && (i_cpg < n_cpg) && (n_new_block_cor  <= min_n_new_block_cor)  ) { // keep adding more CpGs

            for(i=0;i<n_sample;i++) sum_cpg_data[i] += new_cpg_data[i]; // keep track of what is aleady included

            i_cpg_old = i_cpg;
            i_cpg++;
            while ( select_sites[i_cpg]==0 ) {
                for(i=0;i<n_sample;i++) getline(infils[i],STRING);
                i_cpg++;
             }
             for(i=0;i<n_sample;i++) {
   	            getline(infils[i],STRING);
                oldpos = STRING.find(',');
                pos=STRING.find(',',oldpos+1);
                subst = STRING.substr(oldpos+1,pos-oldpos-1);
                new_cpg_data[i] = atof(subst.c_str());
               }

             // calculate information needed to determine is we need to stop
             new_block_cor = calc_mean_cor(cpg_data_matrix, new_cpg_data,block_size,n_sample);

             if(new_block_cor < min_new_block_cor) ++n_new_block_cor ;
                else n_new_block_cor  = 0; // we require 2 adjacent sites to have cor below threshold we reset

             n_cor_double = block_size*(block_size-1)/2;
             weight = n_cor_double / (n_cor_double + block_size);
             prev_mean_cor = mean_cor;
             mean_cor = weight*prev_mean_cor + (1-weight)*new_block_cor;

             for(j=0;j<n_sample;j++) cpg_data_matrix[block_size][j] = new_cpg_data[j];

             prev_block_size = block_size;
             ++block_size;

         } // end while


         // write individual block data
         if (i_cpg == n_cpg)
             {
             if (mean_cor > min_mean_block_cor)
                {
                  block_start_i_cpg[i_block] = block_start;
                  block_end_i_cpg[i_block]   = i_cpg;
                  block_inter_cor[i_block]   = mean_cor;
                  block_size_select[i_block] = block_size;

                  for(i=0;i<n_sample;i++) outfils[i] << i_block+1 << "," << sum_cpg_data[i]/(block_size-1) << "\n";

                  i_block++;
                } // if (mean_cor > min_mean_block_cor)
             else
               if (prev_mean_cor > min_mean_block_cor)
                {
                  block_start_i_cpg[i_block] = block_start;
                  block_end_i_cpg[i_block]   = i_cpg_old;
                  block_inter_cor[i_block]   = prev_mean_cor;
                  block_size_select[i_block] = prev_block_size;

                  for(i=0;i<n_sample;i++) outfils[i] << i_block+1 << "," << sum_cpg_data[i]/(block_size-1) << "\n";

                  i_block++;
                 } // if (prev_mean_cor > min_mean_block_cor)
             } // if (i_cpg == n_cpg)
            else
             if (prev_mean_cor > min_mean_block_cor)
                {
                  block_start_i_cpg[i_block] = block_start;
                  block_end_i_cpg[i_block]   = i_cpg_old;
                  block_inter_cor[i_block]   = prev_mean_cor;
                  block_size_select[i_block] = prev_block_size;

                  for(i=0;i<n_sample;i++) outfils[i] << i_block+1 << "," << sum_cpg_data[i]/(block_size-1) << "\n";

                  i_block++;
                 } // else if (prev_mean_cor > min_mean_block_cor)

            block_start = i_cpg;

    } // end of while (i_cpg < n_cpg)
    for(i=0;i<n_sample;i++) { (infils[i]).close();  (outfils[i]).close(); }


  	ofstream outfile;
    ss << workdir << "blockInfoChr" << ichrom << ".csv";
    filna = ss.str(); ss.str(string()); ptemp= filna.c_str(); outfile.open(ptemp);
    if (!(outfile).is_open()) {cout << " Block info output file " << ptemp << " is not open" << endl; exit(1); }
    cout << " Block info output file is " << filna  <<  "\n";
    outfile << "blocknr,begin,end,all_size,sel_size,meanCor" << "\n"; // Creating header

    for(i=0;i<i_block;i++) outfile << i+1 << "," << coord[block_start_i_cpg[i]] << "," << coord[block_end_i_cpg[i]] << "," << block_end_i_cpg[i]-block_start_i_cpg[i]+1 << "," << block_size_select[i] << "," << block_inter_cor[i] << "\n";

    outfile.close();


  time(&now);
  current = localtime(&now);
  cout << " \n";
  cout << " Finish at " << current->tm_hour << ":" << current->tm_min << ":" << current->tm_sec << "\n";


 //   system("PAUSE");
    return EXIT_SUCCESS;
}



void read_select_sites(const char * filename, int& nr_cpg) {

    string STRING, subst;
	ifstream infile;
	size_t pos, oldpos, pos1, pos2;
	int i=0;

	infile.open (filename);
    if (!(infile).is_open()) { cout << " File with site selection information  " << filename << "  is not open \n"; exit (1); }

	getline(infile,STRING); // Skipping the header

    while(!infile.eof()) {
	        getline(infile,STRING);

	        oldpos = 0;
	        pos=STRING.find(',');
            subst = STRING.substr(oldpos,pos-oldpos);
            coord[i] = atoi(subst.c_str());   // Saving the first in 'coord'

            oldpos = pos+1;
            pos=STRING.find(',',oldpos);
            subst = STRING.substr(oldpos,pos-oldpos);
            select_sites[i] = atoi(subst.c_str());

            i++;
    }

    if (STRING=="") nr_cpg = i-1; else nr_cpg = i;   // make robust to (not) empty final string

    infile.close();

}


void read_first_covfil(const char * filename, int& nr_cpg) {

    string STRING, subst;
	ifstream infile;
	size_t pos, oldpos, pos1, pos2;
	int i=0;

	infile.open (filename);
    if (!(infile).is_open()) { cout << " First coverage file used to determine the number of CpGs " << filename << "  is not open \n"; exit (1); }

	getline(infile,STRING); // Skipping the header

    while(!infile.eof()) {
	        getline(infile,STRING);

	        oldpos = 0;
	        pos=STRING.find(',');
            subst = STRING.substr(oldpos,pos-oldpos);
            coord[i] = atoi(subst.c_str());   // Saving the first in 'coord'

             select_sites[i] = 1;

            i++;
    }

    if (STRING=="") nr_cpg = i-1; else nr_cpg = i;   // make robust to (not) empty final string

    infile.close();

}


void read_sample(const char * filename, int& nr_sample) {

    string STRING;
	ifstream infile;
	int i=0;

	infile.open (filename);

    if (!(infile).is_open()) { cout << " Sample file " << filename << "  is not open \n"; exit (1); }

    while(!infile.eof()) {

	        getline(infile,STRING);
            sample_names[i] = STRING ;
            i++;
    }

    if (STRING == "") nr_sample = i-1; else nr_sample = i; // make robust to (not) empty final string

    infile.close();
}



//---------------------------------------------------

double calc_cor(double * x, double * y, int n)
{
       double xmea=0.0, ymea=0.0;
       int i;

       for(i=0;i<n;i++) xmea +=x[i];
       xmea /= n;

       for(i=0;i<n;i++) ymea +=y[i];
       ymea /= n;

       double numer = 0.0;
       double denom1 = 0.0;
       double denom2 = 0.0;
       double aux = 0.0, bux = 0.0;
       for(i=0;i<n;i++) {
           aux = x[i] - xmea;
           bux = y[i] - ymea;
           numer += aux*bux;
           denom1 += aux*aux;
           denom2 += bux*bux;
       }

       double std1 = sqrt(denom1);
       double std2 = sqrt(denom2);

       double cor;
       if (std1<sd_threshold || std2<sd_threshold) cor=0.0;
        else cor = numer / (std1*std2);

       return (cor);
}

//---------------------------------------------------

double calc_mean_cor(matrix_double_dynamic & X, double * y, int m, int n)
{
       int i,j;

       double * xvec = new double [n];
       double mea = 0;
       for(i=0;i<m;i++) {
           for(j=0;j<n;j++) xvec[j] = X[i][j];
           mea += calc_cor(xvec, y, n);
       }

       mea /= m;
	   cout<<"mean_cor:"<<m << " " <<mea<<endl;
       return (mea);
}

//---------------------------------------------------
void shape_matrix(matrix_double_dynamic & m , int r, int c)
{
  m.clear();
  m.resize(r);
  for (int i=0; i<r; i++)
    m[i].resize(c,0);
}
